# nosec B404
import logging
import subprocess

from behave import given, then, when


@given('I have installed behavex')
def step_impl(context):
    pass  # Assuming behavex is already installed


@when('I run the behavex command')
def step_impl(context):
    context.result = subprocess.run(['behavex', '--help'], capture_output=True, text=True)
    logging.info(context.result.stdout)


@then('I should see the behavex output')
def step_impl(context):
    assert context.result.returncode == 0, "behavex command failed"
    assert "BehaveX - test automation wrapper on top of Behave" in context.result.stdout, "Unexpected output from behavex command"



import logging
import os
import random
import re
import subprocess

from behave import given, then, when

root_project_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
tests_features_path = os.path.join(root_project_path, 'tests', 'features')


@given('The progress bar is enabled')
def step_impl(context):
    context.progress_bar = True


@when('I run the behavex command with a passing test')
@when('I run the behavex command with passing tests')
def step_impl(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'passing_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a file of failing tests')
def step_impl(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', '-rf', os.path.join(tests_features_path, 'failing_scenarios.txt'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command that renames scenarios and features')
def step_impl(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'rename_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a failing test')
def step_impl(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'failing_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a crashing test')
def step_impl(context, parallel_processes="1", parallel_scheme='scenario'):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                os.path.join(tests_features_path, os.path.join(tests_features_path, 'crashing_features', 'crashing_tests.feature')),
                '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a skipped test')
def step_impl(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'skipped_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with an untested test')
def step_impl(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'untested_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_schema}"')
def step_impl(context, parallel_processes, parallel_schema):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features'), '-o', context.output_path, '--parallel-processes', parallel_processes, '--parallel-scheme', parallel_schema]
    execute_command(context, execution_args)


@when('I setup the behavex command with "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_scheme}"')
def step_impl(context, parallel_processes, parallel_scheme):
    context.parallel_processes = parallel_processes
    context.parallel_scheme = parallel_scheme


@when('I run the behavex command with the following scheme, processes and tags')
@when('I run the behavex command with scenario name "{scenario_name}" and the following scheme, processes and tags')
@when('I run the behavex command using "{argument_separator}" separator with the following scheme, processes and tags')
@when('I run the behavex command using "{argument_separator}" separator for "{feature_name}" feature with the following scheme, processes and tags')
@when('I run the behavex command using "{argument_separator}" separator for "{feature_name}" and "{feature_name_2}" features with the following scheme, processes and tags')
def run_command_with_scheme_processes_and_tags(context, scenario_name=None, argument_separator="equal", feature_name=None, feature_name_2=None):
    scheme = context.table[0]['parallel_scheme']
    processes = context.table[0]['parallel_processes']
    tags = context.table[0]['tags']
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    tags_to_folder_name = get_tags_string(tags)
    if not tags:
        tags_array = []
    else:
        tags_array = get_tags_arguments(tags)
    if feature_name:
        if feature_name_2:
            feature_path = os.path.join(tests_features_path, 'secondary_features', feature_name)
            feature_path_2 = os.path.join(tests_features_path, 'secondary_features', feature_name_2)
        else:
            feature_path = os.path.join(tests_features_path, 'secondary_features', feature_name)
    else:
        feature_path = os.path.join(tests_features_path, 'secondary_features')
    if argument_separator == 'equal':
        execution_args = ['behavex', feature_path, '-o', context.output_path, '--parallel-processes=' + processes, '--parallel-scheme=' + scheme] + tags_array
    else:
        execution_args = ['behavex', feature_path, '-o', context.output_path, '--parallel-processes', processes, '--parallel-scheme', scheme] + tags_array
    if feature_name_2:
        # append the second feature path to the execution arguments in index 2
        execution_args.insert(2, feature_path_2)
    if scenario_name:
        execution_args.append('--name')
        execution_args.append(scenario_name.replace(' ', '\\ '))
    execute_command(context, execution_args)


@when('I run the behavex command with the following tags')
def step_impl(context):
    tags = context.table[0]['tags']
    tags_to_folder_name = get_tags_string(tags)
    tags_array = get_tags_arguments(tags)
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features'), '-o', context.output_path] + tags_array
    execute_command(context, execution_args)


@when('I run the behavex command by performing a dry run')
def step_impl(context):
    # generate a random number between 1 and 1000000 completing with zeroes to 6 digits
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features'), '-o', context.output_path, '--dry-run']
    execute_command(context, execution_args)


@then('I should see the following behavex console outputs')
@then('I should see the following behavex console outputs and exit code "{expected_exit_code}"')
def step_impl(context, expected_exit_code=None):
    if expected_exit_code is not None:
        assert int(context.result.returncode) == int(expected_exit_code), "Behavex exit code is not expected"
    for row in context.table:
        assert row['output_line'] in context.result.stdout, f"Unexpected output: {context.result.stdout}\n\nOutput line not found: {row['output_line']}"


@then('I should not see error messages in the output')
def step_impl(context):
    error_messages = ["error", "exception", "traceback"]
    for message in error_messages:
        assert message not in context.result.stdout.lower(), f"Unexpected output: {context.result.stdout}"


@then('I should not see exception messages in the output')
def step_impl(context):
    exception_messages = ["exception", "traceback"]
    for message in exception_messages:
        assert message not in context.result.stdout.lower(), f"Unexpected output: {context.result.stdout}"


@then('I should see the same number of scenarios in the reports and the console output')
def step_impl(context):
    total_scenarios_in_html_report = get_total_scenarios_in_html_report(context)
    logging.info(f"Total scenarios in the HTML report: {total_scenarios_in_html_report}")
    total_scenarios_in_junit_reports = get_total_scenarios_in_junit_reports(context)
    logging.info(f"Total scenarios in the JUnit reports: {total_scenarios_in_junit_reports}")
    total_scenarios_in_console_output = get_total_scenarios_in_console_output(context)
    logging.info(f"Total scenarios in the console output: {total_scenarios_in_console_output}")
    assert total_scenarios_in_html_report == total_scenarios_in_junit_reports == total_scenarios_in_console_output, f"Expected total scenarios to match, but found {total_scenarios_in_html_report} in the HTML report, {total_scenarios_in_junit_reports} in the JUnit reports, and {total_scenarios_in_console_output} in the console output"


@then('I should see the same number of scenarios in the reports')
def verify_total_scenarios_in_reports(context, consider_skipped_scenarios=True):
    total_scenarios_in_html_report = get_total_scenarios_in_html_report(context)
    logging.info(f"Total scenarios in the HTML report: {total_scenarios_in_html_report}")
    total_scenarios_in_junit_reports = get_total_scenarios_in_junit_reports(context, consider_skipped_scenarios)
    logging.info(f"Total scenarios in the JUnit reports: {total_scenarios_in_junit_reports}")
    assert total_scenarios_in_html_report == total_scenarios_in_junit_reports, f"Expected total scenarios to match, but found {total_scenarios_in_html_report} in the HTML report, {total_scenarios_in_junit_reports} in the JUnit reports"


@then('I should see the same number of scenarios in the reports not considering the skipped scenarios')
def step_impl(context):
    verify_total_scenarios_in_reports(context, consider_skipped_scenarios=False)


@then('I should see the HTML report was generated and contains scenarios')
@then('I should see the HTML report was generated and contains "{total_scenarios}" scenarios')
def verify_total_scenarios_in_html_report(context, total_scenarios=None, consider_skipped_scenarios=True):
    total_scenarios_in_html_report = get_total_scenarios_in_html_report(context)
    logging.info(f"Total scenarios in the HTML report: {total_scenarios_in_html_report}")
    if total_scenarios is not None:
        assert total_scenarios_in_html_report == int(total_scenarios), f"Expected the HTML report to contain {total_scenarios} scenarios, but found {total_scenarios_in_html_report}"
    else:
        assert total_scenarios_in_html_report > 0, "Expected the HTML report to be generated and contain scenarios"


@then('I should see the generated HTML report contains the "{string_to_search}" string')
def verify_string_in_html_report(context, string_to_search, string_should_be_present=True):
    total_string_instances_in_html_report = get_string_instances_from_html_report(context, string_to_search)
    logging.info(f"Total instances of '{string_to_search}' in the HTML report: {total_string_instances_in_html_report}")
    if string_should_be_present:
        assert total_string_instances_in_html_report > 0, f"Expected the HTML report to contain the string '{string_to_search}'"
    else:
        assert total_string_instances_in_html_report == 0, f"Expected the HTML report to not contain the string '{string_to_search}'"


@then('I should see the generated HTML report does not contain internal BehaveX variables and tags')
def verify_string_not_in_html_report(context):
    internal_behavex_variables_and_tags = ["BHX_", "BHX_TAG_"]
    for variable_or_tag in internal_behavex_variables_and_tags:
        total_string_instances_in_html_report = get_string_instances_from_html_report(context, variable_or_tag)
        logging.info(f"Total instances of '{variable_or_tag}' in the HTML report: {total_string_instances_in_html_report}")
        assert total_string_instances_in_html_report == 0, f"Expected the HTML report to not contain the string '{variable_or_tag}'"



def get_tags_arguments(tags):
    tags_array = []
    for tag in tags.split(' '):
        tags_array += tag.split('=')
    return tags_array


def get_tags_string(tags):
    return tags.replace('-t=','_AND_').replace('~','NOT_').replace(',','_OR_').replace(' ','').replace('@','')


def get_random_number(total_digits):
    return str(random.randint(1, 1000000)).zfill(total_digits)


def get_total_scenarios_in_console_output(context):
    #Verifying the scenarios in the console output
    console_output = context.result.stdout
    # Extract the number of scenarios by analyzing the following pattern: X scenarios passed, Y failed, Z skipped
    scenario_pattern = re.compile(r'(\d+) scenario.? passed, (\d+) failed, (\d+) skipped')
    match = scenario_pattern.search(console_output)
    if match:
        scenarios_passed = int(match.group(1))
        scenarios_failed = int(match.group(2))
        scenarios_skipped = int(match.group(3))
    else:
        raise ValueError("No scenarios found in the console output")
    return scenarios_passed + scenarios_failed + scenarios_skipped


def get_total_scenarios_in_html_report(context):
    report_path = os.path.abspath(os.path.join(context.output_path, 'report.html'))
    with open(report_path, 'r') as file:
        html_content = file.read()
    return html_content.count('data-scenario-tags=')


def get_string_instances_from_html_report(context, string_to_search):
    report_path = os.path.abspath(os.path.join(context.output_path, 'report.html'))
    with open(report_path, 'r') as file:
        html_content = file.read()
    return html_content.lower().count(string_to_search.lower())


def get_total_scenarios_in_junit_reports(context, consider_skipped_scenarios=True):
    junit_folder = os.path.abspath(os.path.join(context.output_path, 'behave'))
    total_scenarios_in_junit_reports = 0
    for file in os.listdir(junit_folder):
        if file.endswith('.xml'):
            with open(os.path.join(junit_folder, file), 'r') as file:
                xml_content = file.read()
                total_scenarios_in_junit_reports += xml_content.count('status="passed"')
                total_scenarios_in_junit_reports += xml_content.count('status="failed"')
                if consider_skipped_scenarios:
                    total_scenarios_in_junit_reports += xml_content.count('status="skipped"')
    return total_scenarios_in_junit_reports


def execute_command(context, execution_args, print_output=True):
    if "progress_bar" in context and context.progress_bar:
        execution_args.insert(2, '--show-progress-bar')
    if hasattr(context, 'parallel_processes'):
        execution_args += ['--parallel-processes', context.parallel_processes]
    if hasattr(context, 'parallel_scheme'):
        execution_args += ['--parallel-scheme', context.parallel_scheme]
    logging.info("Executing command: {}".format(" ".join(execution_args)))
    context.result = subprocess.run(execution_args, capture_output=True, text=True)
    if print_output:
        logging.info(context.result.stdout)



# Configuration file for the Sphinx documentation builder.

import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------
project = 'BehaveX'
copyright = '2023, hrcorval'
author = 'hrcorval'

# -- General configuration ---------------------------------------------------
extensions = [
    'myst_parser',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
]

# The master toctree document.
master_doc = 'index'

# The suffix(es) of source filenames.
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
html_theme = 'sphinx_rtd_theme'
templates_path = ['_templates']

# -- MyST Parser configuration -----------------------------------------------
myst_enable_extensions = [
    "colon_fence",
]

# -- Setup for markdown files ------------------------------------------------
import commonmark


def setup(app):
    app.connect('html-page-context', add_readme_html)

def add_readme_html(app, pagename, templatename, context, doctree):
    if pagename == 'index':
        with open('../README.md', 'r') as f:
            markdown_content = f.read()
        parser = commonmark.Parser()
        ast = parser.parse(markdown_content)
        renderer = commonmark.HtmlRenderer()
        html = renderer.render(ast)
        context['readme_html'] = html



# -*- coding: utf-8 -*-
"""
/*
* BehaveX - Agile test wrapper on top of Behave (BDD)
*/

This module provides a dictionary with the contents that are displayed in
test execution outputs.
"""

TEXTS = {
    'commons': {
        'expand': 'Expand the {} information',
        'collapse': 'Collapse the {} information',
        'error_background': 'Error in background',
        'framework_name': 'BehaveX',
        'framework_description': 'Agile test wrapper on top of Behave (BDD)',
        'help': {'title': 'Click here for more details'},
        'text': {'total_time': 'Total execution time'},
        'footer': {'name': 'BehaveX'},
    },
    'report': {
        'title': 'Test Report',
        'description': '',
        'modal': {
            'title': '',
            'body': '',
        },
        'muted': 'This scenario has been muted in build server reports (@MUTE tag)',
        'show_background': '(show background)',
        'hide_background': '(hide background)',
        'execution_tag': 'Execution Tag',
        'filter_tag': {'label': 'Tag'},
        'reset_filter': {'label': 'Reset'},
        'skip_fix_process': {'label': 'Skip scenarios under fix process'},
        'icon_duplicate': {'title': 'Copy link to this scenario'},
        'icon_repeat': {
            'title': 'This scenario was executed more than once (see @AUTORETRY tag)'
        },
        'filter_status': {'label': 'Status'},
    },
    'steps': {'title': 'Steps', 'description': '', 'modal': {'title': '', 'body': ''}},
    'metrics': {
        'title': 'Metrics',
        'description': '',
        'modal': {
            'title': '',
            'body': '<b>Test Automation Rate:</b> % of automated scenarios.<br>'
            '<b>Pass Rate:</b>: % of passed scenarios.<br>',
        },
    },
    'joined': {'title': '', 'description': '', 'modal': {'title': '', 'modal': 'the'}},
    'feature': {
        'serial_execution': '\n{0}\nRunning serial features (tagged as @SERIAL).\n{0}\n'.format('*' * 60),
        'running_parallels': '\n{0}\nRunning parallel features.\n{0}\n'.format('*' * 60),
        'empty_scenario_descriptions': u'{0}\nThere are features containing scenarios with empty descriptions: \n* {1}.\n{0}\n'.format('*' * 60, {}),
        'run_behave': u"Running feature '{}'.",
        'execution_crashed': 'Execution crashed or was interrupted when executing the feature. No outputs could be generated.'
    },
    'scenario': {
        'serial_execution': u'\n{0}\nRunning serial scenarios (tagged as @SERIAL).\n{0}\n'.format('*' * 60),
        'running_parallels': u'\n{0}\nRunning parallel scenarios\n{0}\n'.format('*' * 60),
        'run_behave': u"Running feature '{}' with scenario '{}'.",
        'duplicated_scenarios': '{0}\nThere are duplicate scenario names to run.\n'
        'Parallel test execution by scenario cannot be performed.\n'
        'Duplicated scenario names: \n{1}.\n{0}\n'.format('*' * 60, {}),
        'execution_crashed': 'Execution crashed or was interrupted when executing the scenario. No outputs could be generated.'
    },
    'folder': {'run_behave': u"Running folder: '{}' and feature '{}'."},
    'path': {'not_found': u'\nThe path "{}" was not found.\n'},
}



# -*- coding: utf-8 -*-
"""
/*
* BehaveX - Agile test wrapper on top of Behave (BDD)
*/
"""
# __future__ has been added in order to maintain compatibility
from __future__ import absolute_import

import os
import re

from behave.model_core import Status

from behavex.conf_mgr import get_env
from behavex.global_vars import global_vars
from behavex.outputs.jinja_mgr import TemplateHandler
from behavex.outputs.report_utils import (get_save_function,
                                          match_for_execution,
                                          retry_file_operation, text)
from behavex.utils import get_scenario_tags


def _export_feature_to_xml(feature, isobject=True):
    def get_scenarios(feature_):
        def flatter_scenarios(scenarios_list):
            return sum(
                (
                    [scenario] if scenario.type == 'scenario' else scenario._scenarios
                    for scenario in scenarios_list
                ),
                [],
            )
        scenarios = flatter_scenarios(feature_.scenarios) if isobject else feature_['scenarios']
        return scenarios

    def get_status(scenario_):
        if hasattr(scenario_, 'status'):
            return scenario_.status
        else:
            status = scenario_['status']
            if 'untested' in status:
                return Status.untested
            elif 'failed' in status:
                return Status.failed
            elif 'skipped' in status:
                return Status.skipped
            elif 'passed' in status:
                return Status.passed

    scenarios = [
        scenario
        for scenario in get_scenarios(feature)
        if (match_for_execution(get_scenario_tags(scenario)) and not (get_status(scenario) == Status.skipped and get_env('RERUN_FAILURES')))
    ]

    skipped = [scenario for scenario in scenarios if get_status(scenario) == Status.skipped]
    failures = [
        scenario
        for scenario in scenarios
        if get_status(scenario) == Status.failed or get_status(scenario) == Status.untested
    ]
    muted = [
        scenario
        for scenario in scenarios
        if any(i in ['MUTE'] for i in get_scenario_tags(scenario))
    ]

    muted_failed = [scenario for scenario in muted if get_status(scenario) == Status.failed]

    summary = {
        'time': sum(
            scenario.duration if isobject else scenario['duration']
            for scenario in scenarios
        ),
        'tests': len(scenarios),
        'failures': len(failures) - len(muted_failed),
        'skipped': len(skipped) + len(muted_failed),
    }
    parameters_template = {
        'feature': feature,
        'summary': summary,
        'skipped': skipped,
        'failures': failures,
        'scenarios': scenarios,
        'muted': muted_failed,
    }

    template_handler = TemplateHandler(global_vars.jinja_templates_path)
    output_text = template_handler.render_template(
        global_vars.jinja_templates['xml']
        if isobject
        else global_vars.jinja_templates['xml_json'],
        parameters_template,
    )
    output_text = output_text.replace('status="Status.', 'status="')
    exp = re.compile('\\\\|/')
    filename = feature.filename if isobject else feature['filename']
    filename = text(filename)
    name = u'.'.join(exp.split(filename.partition('features')[2][1:-8]))
    junit_path = os.path.join(get_env('OUTPUT'), 'behave')
    path_output = os.path.join(junit_path, u'TESTS-' + name + u'.xml')

    retry_file_operation(
        path_output + '.xml', get_save_function(path_output, output_text)
    )


def export_feature_to_xml(feature, isobject=True):
    _export_feature_to_xml(feature, isobject)



import platform
import sys
import time

from behavex.global_vars import global_vars


def get_progress_chars():
    """Get appropriate progress bar characters based on platform."""
    if platform.system() == 'Windows':
        return {
            'bar': '#',
            'edge': '|',
            'empty': '-'
        }
    return {
        'bar': '█',
        'edge': '|',
        'empty': '-'
    }


class ProgressBar:
    def __init__(self, prefix, total, print_updates_in_new_lines=False):
        self.prefix = prefix
        self.total = total
        self.print_updates_in_new_lines = print_updates_in_new_lines
        self.progress_chars = get_progress_chars()
        self.bar_length = 15
        self.current_iteration = 0
        self.start_time = global_vars.execution_start_time

    def start(self, start_increment=0):
        self.current_iteration = start_increment
        self._print_progress_bar(new_line=True)

    def update(self, increment=1):
        self.current_iteration += increment
        self._print_progress_bar()

    def finish(self, print_if_total_reached=False):
        if print_if_total_reached or self.current_iteration < self.total:
            self.current_iteration = self.total
            self._print_progress_bar(new_line=True)

    def _print_progress_bar(self, new_line=False):
        prefix = f"{self.prefix}: " if self.prefix else ""
        if self.total == 0:
            percent = 100
            filled_length = int(self.bar_length)
        else:
            percent = 100 * float(self.current_iteration / float(self.total))
            filled_length = int(self.bar_length * self.current_iteration // self.total)
        bar = self.progress_chars['bar'] * filled_length + self.progress_chars['empty'] * (self.bar_length - filled_length)
        elapsed_time = global_vars.execution_elapsed_time
        elapsed_formatted = time.strftime("%M:%S", time.gmtime(elapsed_time))
        progress_bar_content = f"\r{prefix}{percent:.0f}%|{bar}| {self.current_iteration}/{self.total} [{elapsed_formatted}]\r"
        if self.print_updates_in_new_lines or new_line or percent == 100:
            print(progress_bar_content)
        else:
            sys.stdout.write(progress_bar_content)
            sys.stdout.flush()



# -*- coding: utf-8 -*-

"""BehaveX - Agile test wrapper on top of Behave (BDD).

This module provides the main entry point for running BehaveX tests,
including setup, execution, and reporting.
"""

# pylint: disable=W0703

# __future__ has been added to maintain compatibility
from __future__ import absolute_import, print_function

import codecs
import copy
import json
import logging.config
import multiprocessing
import os
import os.path
import platform
import re
import signal
import sys
import time
import traceback
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import active_children
from multiprocessing.managers import DictProxy
from tempfile import gettempdir

from behave import __main__ as behave_script
from behave.model import Feature, Scenario, ScenarioOutline

# noinspection PyUnresolvedReferences
import behavex.outputs.report_json
from behavex import conf_mgr
from behavex.arguments import BEHAVE_ARGS, BEHAVEX_ARGS, parse_arguments
from behavex.conf_mgr import ConfigRun, get_env, get_param, set_env
from behavex.environment import extend_behave_hooks
from behavex.execution_singleton import ExecutionSingleton
from behavex.global_vars import global_vars
from behavex.outputs import report_xml
from behavex.outputs.report_json import generate_execution_info
from behavex.outputs.report_utils import (get_overall_status,
                                          match_for_execution,
                                          pretty_print_time,
                                          retry_file_operation, text)
from behavex.progress_bar import ProgressBar
from behavex.utils import (IncludeNameMatch, IncludePathsMatch, MatchInclude,
                           cleanup_folders, configure_logging,
                           copy_bootstrap_html_generator,
                           create_execution_complete_callback_function,
                           expand_paths, explore_features, generate_hash,
                           generate_reports, get_feature_and_scenario_line,
                           get_json_results, get_logging_level,
                           get_scenario_tags, get_scenarios_instances,
                           get_text, join_feature_reports,
                           join_scenario_reports, len_scenarios,
                           print_env_variables, print_parallel,
                           set_behave_tags, set_env_variable,
                           set_environ_config, set_system_paths)

EXIT_OK = 0
EXIT_ERROR = 1
EXECUTION_BLOCKED_MSG = (
    'Some of the folders or files are being used by another '
    'program. Please, close them and try again...'
)

os.environ.setdefault('EXECUTION_CODE', '1')
match_include = None
include_path_match = None
include_name_match = None


def main():
    """BehaveX starting point.

    Parses command-line arguments and initiates the test run.
    """
    args = sys.argv[1:]
    exit_code = run(args)
    try:
        exit(exit_code)
    except:
        # force exit
        sys.exit(exit_code)

def run(args):
    """Run BehaveX with the given arguments.

    Args:
        args (list): Command-line arguments.

    Returns:
        int: Exit code indicating success or failure.
    """
    global match_include
    global include_path_match
    global include_name_match
    args_parsed = parse_arguments(args)
    set_environ_config(args_parsed)
    ConfigRun().set_args(args_parsed)
    execution_code, rerun_list = setup_running_failures(args_parsed)
    if rerun_list:
        paths = ",".join(rerun_list).replace('\n', '')
        os.environ['FEATURES_PATH'] = paths
        global_vars.rerun_failures = True
        if execution_code == EXIT_ERROR:
            return EXIT_ERROR
    else:
        # Handle paths parameter
        if len(get_param('paths')) > 0:
            paths = ",".join(expand_paths(get_param('paths'))).replace('\n', '')
            os.environ['FEATURES_PATH'] = paths

        # Handle include_paths parameter
        if len(get_param('include_paths')) > 0:
            include_paths = ",".join(expand_paths(get_param('include_paths'))).replace('\n', '')
            features_path = os.environ.get('FEATURES_PATH')
            if features_path == '' or features_path is None or not os.path.exists(features_path):
                os.environ['FEATURES_PATH'] = include_paths
            else:
                os.environ['FEATURES_PATH'] = features_path + ',' + include_paths
    features_path = os.environ.get('FEATURES_PATH')
    if features_path == '' or features_path is None:
        os.environ['FEATURES_PATH'] = 'features'
    _set_env_variables(args_parsed)
    set_system_paths()
    cleanup_folders()
    copy_bootstrap_html_generator()
    configure_logging(args_parsed)
    match_include = MatchInclude()
    include_path_match = IncludePathsMatch()
    include_name_match = IncludeNameMatch()

    return launch_behavex()


def setup_running_failures(args_parsed):
    """Setup the environment for rerunning failed tests.

    Args:
        args_parsed (Namespace): Parsed command-line arguments.

    Returns:
        tuple: Exit code and list of paths to rerun.
    """
    failures_path = args_parsed.rerun_failures
    if failures_path:
        failures_path = os.path.normpath(failures_path)
        if not os.path.isabs(failures_path):
            failures_path = os.path.abspath(failures_path)
        if os.path.exists(failures_path):
            set_env_variable('RERUN_FAILURES', args_parsed.rerun_failures)
            with open(failures_path, 'r') as failures_file:
                content = failures_file.read()
                if not content:
                    print('\nThere are no failing test scenarios to run.')
                    return EXIT_ERROR, None
                return EXIT_OK, content.split(",")
        else:
            print('\nThe specified failing scenarios filename was not found: {}'.format(failures_path))
            return EXIT_ERROR, None
    else:
        return EXIT_OK, None


def init_multiprocessing(idQueue, parallel_delay):
    try:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # Retrieve one of the unique IDs
        worker_id = idQueue.get()
        # Use the unique ID to name the process
        multiprocessing.current_process().name = f'behave_worker-{worker_id}'
        # Add an initial delay to avoid all processes starting at the same time
        if isinstance(parallel_delay, int) and parallel_delay > 0:
            time.sleep(parallel_delay * worker_id / 1000.0)
    except Exception as e:
        logging.error(f"Exception in init_multiprocessing: {e}")
        raise


def launch_behavex():
    """Launch the BehaveX test execution in the specified parallel mode.

    Returns:
        int: Exit code indicating success or failure.
    """
    json_reports = []
    execution_codes = []
    results = None
    config = conf_mgr.get_config()
    features_path = os.environ.get('FEATURES_PATH')
    parallel_scheme = get_param('parallel_scheme')
    if get_param('dry_run'):
        parallel_processes = 1
        show_progress_bar = False
    else:
        parallel_processes = get_param('parallel_processes')
        parallel_processes = 1 if parallel_processes <= 1 else parallel_processes
        show_progress_bar = get_param('show_progress_bar')
    multiprocess = (
        True
        if get_param('parallel_processes') > 1 and not get_param('dry_run')
        else False
    )
    set_behave_tags()
    scenario = False
    notify_missing_features(features_path)
    features_list = {}
    for path in features_path.split(','):
        features_list[path] = explore_features(path)
    updated_features_list = create_scenario_line_references(features_list)
    parallel_scheme = '' if not multiprocess else parallel_scheme
    manager = multiprocessing.Manager()
    # shared variable to track scenarios that should be run but seems to be removed from execution (using scenarios.remove)
    shared_removed_scenarios = manager.dict()
    lock = manager.Lock()
    # Create a queue containing unique IDs from 0 to the number of parallel processes - 1
    # These IDs will be attributed to the process when they will be initialized
    idQueue = manager.Queue()
    for i in range(parallel_processes):
        idQueue.put(i)
    parallel_delay = get_param('parallel_delay')
    process_pool = ProcessPoolExecutor(max_workers=parallel_processes,
                                       initializer=init_multiprocessing,
                                       initargs=(idQueue, parallel_delay))
    global_vars.execution_start_time = time.time()
    totals = {"features": {"passed": 0, "failed": 0, "skipped": 0, "untested": 0},
            "scenarios": {"passed": 0, "failed": 0, "skipped": 0, "untested": 0}}
    try:
        config = ConfigRun()
        if parallel_processes == 1 or get_param('dry_run'):
            # Executing without parallel processes
            if get_param('dry_run'):
                print('Obtaining information about the reporting scope...')
            if global_vars.rerun_failures:
                all_paths = features_path.split(",")
            else:
                all_paths = [key for key in updated_features_list]
            if len(all_paths) > 0:
                execution_codes, json_reports = execute_tests(features_path=all_paths,
                                                                feature_filename=None,
                                                                feature_json_skeleton=None,
                                                                scenarios_to_run_in_feature=None,
                                                                scenario_line=None,
                                                                multiprocess=False,
                                                                config=config,
                                                                lock=None,
                                                                shared_removed_scenarios=None)
            else:
                execution_codes, json_reports = (0, [{'environment': [], 'features': [], 'steps_definition': []}])
        elif parallel_scheme == 'scenario':
            execution_codes, json_reports = launch_by_scenario(updated_features_list,
                                                            process_pool,
                                                            lock,
                                                            shared_removed_scenarios,
                                                            show_progress_bar)
            scenario = True
        elif parallel_scheme == 'feature':
            execution_codes, json_reports = launch_by_feature(updated_features_list,
                                                            process_pool,
                                                            lock,
                                                            show_progress_bar)
        wrap_up_process_pools(process_pool, json_reports, scenario)
        time_end = global_vars.execution_end_time

        if get_param('dry_run'):
            msg = '\nDry run completed. Please, see the report in {0}' ' folder.\n\n'
            print(msg.format(get_env('OUTPUT')))

        remove_temporary_files(parallel_processes, json_reports)

        failing_non_muted_tests = False
        # TODO: Replace logs below with test execution logs when an unexpected error occurs
        # behave_log_file = os.path.join(output_folder, 'behavex', 'logs', str(scenario['id_feature']), 'behave.log')
        # behave_log_file = os.path.join(output_folder, 'behavex', 'logs', str(json_test_configuration['id']), 'behave.log')
        results = get_json_results()
        processed_feature_filenames = []
        if results:
            failures = []
            for feature in results['features']:
                processed_feature_filenames.append(feature['filename'])
                filename = feature['filename']
                if feature['status'] == 'failed':
                    totals['features']['failed'] += 1
                elif feature['status'] == 'passed':
                    totals['features']['passed'] += 1
                elif feature['status'] == 'untested':
                    totals['features']['untested'] += 1
                else:
                    totals['features']['skipped'] += 1
                    totals['scenarios']['skipped'] += len(feature['scenarios'])
                    continue
                for scenario in feature['scenarios']:
                    if scenario['status'] == 'failed':
                        totals['scenarios']['failed'] += 1
                        failures.append('{}:{}'.format(filename, scenario['line']))
                        if 'MUTE' not in scenario['tags']:
                            failing_non_muted_tests = True
                    elif scenario['status'] == 'passed':
                        totals['scenarios']['passed'] += 1
                    elif scenario['status'] == 'untested':
                        totals['scenarios']['untested'] += 1
                    else:
                        totals['scenarios']['skipped'] += 1
            if failures:
                failures_file_path = os.path.join(get_env('OUTPUT'), global_vars.report_filenames['report_failures'])
                with open(failures_file_path, 'w') as failures_file:
                    failures_file.write(','.join(failures))
        # Calculates final exit code. execution_codes is 1 only if an execution exception arises
        if isinstance(execution_codes, list):
            execution_failed = True if sum(execution_codes) > 0 else False
            execution_interrupted_or_crashed = True if any([code == 2 for code in execution_codes]) else False
        else:
            execution_failed = True if execution_codes > 0 else False
            execution_interrupted_or_crashed = True if execution_codes == 2 else False
        exit_code = (EXIT_ERROR if (execution_failed and failing_non_muted_tests) or execution_interrupted_or_crashed else EXIT_OK)
    except KeyboardInterrupt as ex:
        print('Caught KeyboardInterrupt, terminating workers')
        try:
            while active_children():
                for process in active_children():  # Terminate all active child processes
                    process.terminate()  # Forcefully terminate each process
                time.sleep(1)
            process_pool.shutdown(wait=False)
        except Exception as e:
            print(f"Error during shutdown: {e}")
        exit_code = EXIT_ERROR
    if multiprocess:
        untested_features = totals['features']['untested']
        untested_scenarios = totals['scenarios']['untested']
        untested_features_msg = ', {} untested'.format(untested_features) if untested_features > 0 else ''
        untested_scenarios_msg = ', {} untested'.format(untested_scenarios) if untested_scenarios > 0 else ''
        def plural_char(n): return 's' if n != 1 else ''
        print('\n{} feature{} passed, {} failed, {} skipped{}'.format(totals['features']['passed'],
                                                                        plural_char(totals['features']['passed']),
                                                                        totals['features']['failed'],
                                                                        totals['features']['skipped'],
                                                                        untested_features_msg))
        print('{} scenario{} passed, {} failed, {} skipped{}'.format(totals['scenarios']['passed'],
                                                                       plural_char(totals['scenarios']['passed']),
                                                                       totals['scenarios']['failed'],
                                                                       totals['scenarios']['skipped'],
                                                                       untested_scenarios_msg))
        # TODO: print steps execution summary ('{} steps passed, {} failed, {} skipped{}, {} untested')
        print('Took: {}'.format(pretty_print_time(global_vars.execution_end_time - global_vars.execution_start_time)))
    if results and results['features']:
        print('\nHTML output report is located at: {}'.format(os.path.join(get_env('OUTPUT'), "report.html")))
    print('Exit code: {}'.format(exit_code))
    return exit_code


def notify_missing_features(features_path):
    """Notify if any features are missing in the specified path.

    Args:
        features_path (str): Path to the features.
    """
    if global_vars.rerun_failures:
        all_paths = features_path.split(",")
        for path in all_paths:
            include_path = path.partition(':')[0]
            if not os.path.exists(os.path.normpath(include_path)):
                print_parallel('path.not_found', os.path.realpath(include_path))


def create_scenario_line_references(features):
    """Create references for scenario lines in the features.

    Args:
        features (dict): Dictionary of features and their scenarios.

    Returns:
        dict: Updated features with scenario line references.
    """
    updated_features = {}
    for feature_path, scenarios in features.items():
        for scenario in scenarios:
            scenario_filename = text(scenario.filename)
            if global_vars.rerun_failures or ".feature:" in feature_path:
                if feature_path not in updated_features:
                    updated_features[feature_path] = []
                if isinstance(scenario, ScenarioOutline):
                    for scenario_outline_instance in scenario.scenarios:
                        outline_scenario_line = get_feature_and_scenario_line(scenario_outline_instance.name)[1]
                        if outline_scenario_line and scenario_outline_instance.line == int(outline_scenario_line):
                            if scenario_outline_instance not in updated_features[pure_feature_path]:
                                updated_features[feature_path].append(scenario_outline_instance)
                            break
                else:
                    scenario_line = get_feature_and_scenario_line(feature_path)[1]
                    if scenario_line and scenario.line == int(scenario_line):
                        if scenario not in updated_features[feature_path]:
                            updated_features[feature_path].append(scenario)
            else:
                updated_features_path = scenario.feature.filename
                if updated_features_path not in updated_features:
                    updated_features[updated_features_path] = []
                if isinstance(scenario, ScenarioOutline):
                    for scenario_outline_instance in scenario.scenarios:
                        if scenario_outline_instance not in updated_features[updated_features_path]:
                            updated_features[updated_features_path].append(scenario_outline_instance)
                else:
                    if scenario not in updated_features[updated_features_path]:
                        updated_features[updated_features_path].append(scenario)
    return updated_features


def launch_by_feature(features,
                      process_pool,
                      lock,
                      show_progress_bar):
    """Launch tests by feature in parallel.

    Args:
        features (dict): Dictionary of features and their scenarios.
        process_pool (ProcessPoolExecutor): Process pool executor.
        lock (Lock): Multiprocessing lock.
        show_progress_bar (bool): Whether to show the progress bar.

    Returns:
        tuple: Execution codes and JSON reports.
    """
    json_reports = []
    execution_codes = []
    serial_features = []
    parallel_features = []
    for features_path in features:
        feature = features[features_path][0].feature
        pure_feature_path, scenario_line = get_feature_and_scenario_line(features_path)
        feature_filename = feature.filename if not scenario_line else "{}:{}".format(pure_feature_path, scenario_line)
        if 'SERIAL' in feature.tags:
            serial_features.append({"feature_filename": feature_filename,
                                    "feature_json_skeleton": _get_feature_json_skeleton(feature)})
        else:
            parallel_features.append({"feature_filename": feature_filename,
                                      "feature_json_skeleton": _get_feature_json_skeleton(feature)})
    if show_progress_bar:
        total_features = len(serial_features) + len(parallel_features)
        global_vars.progress_bar_instance = _get_progress_bar_instance(parallel_scheme="feature",
                                                                       total_elements=total_features)
        if global_vars.progress_bar_instance:
            global_vars.progress_bar_instance.start()
    if serial_features:
        print_parallel('feature.serial_execution')
        for serial_feature in serial_features:
            execution_code, map_json = execute_tests(features_path=None,
                                                     feature_filename=serial_feature["feature_filename"],
                                                     feature_json_skeleton=serial_feature["feature_json_skeleton"],
                                                     scenarios_to_run_in_feature=None,
                                                     scenario_line=None,
                                                     multiprocess=True,
                                                     config=ConfigRun(),
                                                     lock=None,
                                                     shared_removed_scenarios=None)
            json_reports += [map_json]
            execution_codes.append(execution_code)
            if global_vars.progress_bar_instance:
                global_vars.progress_bar_instance.update()
    print_parallel('feature.running_parallels')
    parallel_processes = []
    for parallel_feature in parallel_features:
        feature_filename = parallel_feature["feature_filename"]
        feature_json_skeleton = parallel_feature["feature_json_skeleton"]
        future = process_pool.submit(execute_tests,
                                     features_path=None,
                                     feature_filename=feature_filename,
                                     feature_json_skeleton=feature_json_skeleton,
                                     scenarios_to_run_in_feature=None,
                                     scenario_line=None,
                                     multiprocess=True,
                                     config=ConfigRun(),
                                     lock=lock,
                                     shared_removed_scenarios=None)
        parallel_processes.append(future)
        future.add_done_callback(create_execution_complete_callback_function(
            execution_codes,
            json_reports,
            global_vars.progress_bar_instance,
        ))
    for parallel_process in parallel_processes:
        parallel_process.result()
    return execution_codes, json_reports


def launch_by_scenario(features,
                       process_pool,
                       lock,
                       shared_removed_scenarios,
                       show_progress_bar):
    """Launch tests by scenario in parallel.

    Args:
        features (dict): Dictionary of features and their scenarios.
        process_pool (ProcessPoolExecutor): Process pool executor.
        lock (Lock): Multiprocessing lock.
        shared_removed_scenarios (dict): Shared dictionary of removed scenarios.
        show_progress_bar (bool): Whether to show the progress bar.

    Returns:
        tuple: Execution codes and JSON reports.
    """
    json_reports = []
    execution_codes = []
    parallel_scenarios = {}
    serial_scenarios = {}
    duplicated_scenarios = {}
    total_scenarios_to_run = {}
    features_with_no_scen_desc = []
    for features_path, scenarios in features.items():
        scenarios_instances = get_scenarios_instances(scenarios)
        for scenario in scenarios_instances:
            if include_path_match(scenario.filename, scenario.line) \
                    and include_name_match(scenario.name):
                scenario_tags = get_scenario_tags(scenario)
                if match_for_execution(scenario_tags):
                    if scenario.name == "":
                        features_with_no_scen_desc.append(scenario.filename)
                    feature_json_skeleton = _get_feature_json_skeleton(scenario)
                    feature_filename = scenario.feature.filename
                    scenario_information = {"feature_filename": feature_filename,
                                            "feature_json_skeleton": feature_json_skeleton,
                                            "scenario_line": scenario.line}
                    total_scenarios_to_run[feature_filename] = total_scenarios_to_run.setdefault(feature_filename, 0) + 1
                    if 'SERIAL' in scenario_tags:
                        for key, list_scenarios in serial_scenarios.items():
                            if scenario_information in list_scenarios:
                                duplicated_scenarios.setdefault(key, []).append(scenario.name)
                        serial_scenarios.setdefault(features_path, []).append(scenario_information)
                    else:
                        for key, list_scenarios in parallel_scenarios.items():
                            if scenario_information in list_scenarios:
                                duplicated_scenarios.setdefault(key, []).append(scenario.name)
                        parallel_scenarios.setdefault(features_path, []).append(scenario_information)
    if show_progress_bar:
        global_vars.progress_bar_instance = _get_progress_bar_instance(parallel_scheme="scenario",
                                                                       total_elements=sum(total_scenarios_to_run.values()))
        if global_vars.progress_bar_instance:
            global_vars.progress_bar_instance.start()
    if duplicated_scenarios:
        print_parallel('scenario.duplicated_scenarios',
                       json.dumps(duplicated_scenarios, indent=4))
        exit(1)
    if features_with_no_scen_desc:
        print_parallel('feature.empty_scenario_descriptions',
                       '\n* '.join(features_with_no_scen_desc))
        exit(1)
    if serial_scenarios:
        print_parallel('scenario.serial_execution')
        for features_path, scenarios_in_feature in serial_scenarios.items():
            for scen_info in scenarios_in_feature:
                scenarios_to_run_in_feature = total_scenarios_to_run[scen_info["feature_filename"]]
                execution_code, json_report = execute_tests(features_path=features_path,
                                                            feature_filename=scen_info["feature_filename"],
                                                            feature_json_skeleton=scen_info["feature_json_skeleton"],
                                                            scenarios_to_run_in_feature=scenarios_to_run_in_feature,
                                                            scenario_line=scen_info["scenario_line"],
                                                            multiprocess=True,
                                                            config=ConfigRun(),
                                                            lock=None,
                                                            shared_removed_scenarios=shared_removed_scenarios)
                execution_codes.append(execution_code)
                json_reports.append(json_report)
                if global_vars.progress_bar_instance:
                    global_vars.progress_bar_instance.update()
    if parallel_scenarios:
        print_parallel('scenario.running_parallels')
        parallel_processes = []
        for features_path in parallel_scenarios.keys():
            for scenario_information in parallel_scenarios[features_path]:
                scenarios_to_run_in_feature = total_scenarios_to_run[scenario_information["feature_filename"]]
                feature_filename = scenario_information["feature_filename"]
                feature_json_skeleton = scenario_information["feature_json_skeleton"]
                scenario_line = scenario_information["scenario_line"]
                future = process_pool.submit(execute_tests,
                                             features_path=features_path,
                                             feature_filename=feature_filename,
                                             feature_json_skeleton=feature_json_skeleton,
                                             scenarios_to_run_in_feature=scenarios_to_run_in_feature,
                                             scenario_line=scenario_line,
                                             multiprocess=True,
                                             config=ConfigRun(),
                                             lock=lock,
                                             shared_removed_scenarios=shared_removed_scenarios
                                             )
                parallel_processes.append(future)
                future.add_done_callback(create_execution_complete_callback_function(
                    execution_codes,
                    json_reports,
                    global_vars.progress_bar_instance
                ))
        for parallel_process in parallel_processes:
            parallel_process.result()
    return execution_codes, json_reports


def execute_tests(
        features_path,
        feature_filename,
        feature_json_skeleton,
        scenarios_to_run_in_feature,
        scenario_line,
        multiprocess,
        config,
        lock,
        shared_removed_scenarios):
    """
    Execute tests for the given feature or scenario.

    Args:
        features_path (str): Path to the features.
        feature_filename (str): Name of the feature file.
        feature_json_skeleton (str): JSON skeleton of the feature.
        scenarios_to_run_in_feature (int): Number of scenarios to run in the feature.
        scenario_line (int): Line of the scenario.
        multiprocess (bool): Whether to use multiprocessing.
        config (ConfigRun): Configuration object.
        lock (Lock): Multiprocessing lock.
        shared_removed_scenarios (dict): Shared dictionary of removed scenarios.

    Returns:
        tuple: Execution code and JSON report.
    """
    try:
        behave_args = None
        if multiprocess:
            ExecutionSingleton._instances[ConfigRun] = config
        extend_behave_hooks()
        try:
            # Execution ID is only important for multiprocessing so that
            # we can influence where output files end up
            execution_id = json.loads(feature_json_skeleton or '{}').get('id')
            behave_args = _set_behave_arguments(features_path=features_path,
                                                multiprocess=multiprocess,
                                                execution_id=execution_id,
                                                feature=feature_filename,
                                                scenario_line=scenario_line,
                                                config=config)
        except Exception as exception:
            traceback.print_exc()
            print(exception)
        execution_code, generate_report = _launch_behave(behave_args)
        # print("pipenv run behave {} --> Execution Code: {} --> Generate Report: {}".format(" ".join(behave_args), execution_code, generate_report))
        if generate_report:
            # print execution code
            if execution_code == 2:
                if feature_json_skeleton:
                    json_output = {'environment': [],
                                   'features': [json.loads(feature_json_skeleton)],
                                   'steps_definition': []}
                    for skeleton_feature in json_output["features"]:
                        if scenario_line:
                            for skeleton_scenario in skeleton_feature["scenarios"]:
                                if str(skeleton_scenario['line']) == str(scenario_line):
                                    skeleton_scenario['status'] = 'failed'
                                    skeleton_scenario['error_msg'] = get_text('scenario.execution_crashed')
                        else:
                            skeleton_feature['status'] = 'failed'
                            skeleton_feature['error_msg'] = 'Execution crashed. No outputs could be generated.'
                            for skeleton_scenario in skeleton_feature["scenarios"]:
                                skeleton_scenario['status'] = 'failed'
                                skeleton_scenario['error_msg'] = get_text('feature.execution_crashed')
                else:
                    json_output = {'environment': [], 'features': [], 'steps_definition': []}
            else:
                json_output = dump_json_results()
            if scenario_line:
                json_output['features'] = filter_feature_executed(json_output,
                                                                  text(feature_filename),
                                                                  scenario_line=scenario_line)
                if len(json_output['features']) == 0 or len(json_output['features'][0]['scenarios']) == 0:
                    # Adding scenario data if the test was removed from the execution (setting it as "Untested")
                    json_output['features'] = [json.loads(feature_json_skeleton)]
                try:
                    processing_xml_feature(json_output=json_output,
                                           scenario_line=scenario_line,
                                           feature_filename=feature_filename,
                                           scenarios_to_run_in_feature=scenarios_to_run_in_feature,
                                           lock=lock,
                                           shared_removed_scenarios=shared_removed_scenarios)
                except Exception as ex:
                    logging.exception("There was a problem processing the xml file: {}".format(ex))
        else:
            json_output = {'environment': [], 'features': [], 'steps_definition': []}
        return execution_code, join_feature_reports(json_output)
    except Exception as e:
        logging.error(f"Exception in execute_tests: {e}")
        raise


def filter_feature_executed(json_output, filename, scenario_line):
    """
    Filter the executed feature from the JSON output.

    Args:
        json_output (dict): JSON output of the test execution.
        filename (str): Name of the feature file.
        scenario_line (str): Line of the scenario.
    """
    for feature in json_output.get('features', '')[:]:
        if feature.get('filename', '') == filename:
            mapping_scenarios = []
            for scenario in feature['scenarios']:
                if str(scenario['line']) == str(scenario_line):
                    mapping_scenarios.append(scenario)
            feature['scenarios'] = mapping_scenarios
            return [feature]
    return []


def _launch_behave(behave_args):
    """
    Launch Behave with the given arguments.

    Args:
        behave_args (list): List of arguments for Behave.

    Returns:
        tuple: Execution code and whether to generate a report.
    """
    generate_report = True
    execution_code = 0
    stdout_file = None

    try:
        stdout_file = behave_args[behave_args.index('--outfile') + 1]
        execution_code = behave_script.main(behave_args)
        if not os.path.exists(stdout_file):
            execution_code = 2
            generate_report = True

    except KeyboardInterrupt:
        execution_code = 1
        generate_report = False
    except Exception as ex:
        execution_code = 1
        generate_report = True
        logging.exception('Unexpected error executing behave steps: ')
        logging.exception(ex)
        traceback.print_exc()
    except:
        execution_code = 2
        generate_report = True

    if stdout_file and os.path.exists(stdout_file):
        def _merge_and_remove():
            with open(os.path.join(get_env('OUTPUT'), 'merged_behave_outputs.log'), 'a+') as behave_log_file:
                with open(stdout_file, 'r') as source_file:
                    behave_log_file.write(source_file.read())
            # nosec B110
            try:
                if not source_file.closed:
                    os.close(source_file.fileno())
                if os.path.exists(stdout_file):
                    os.remove(stdout_file)
            except:
                pass
        try:
            retry_file_operation(stdout_file, _merge_and_remove)
        except Exception as remove_ex:
            logging.warning(f"Could not remove stdout file {stdout_file}: {remove_ex}")
            # Don't fail the execution if we can't remove the file
            pass

    return execution_code, generate_report


def wrap_up_process_pools(process_pool,
                          json_reports,
                          scenario=False):
    """
    Wrap up the process pools and generate the final report.

    Args:
        process_pool (ProcessPoolExecutor): Process pool executor.
        json_reports (list): List of JSON reports.
        scenario (bool): Whether the execution was by scenario.
    """
    merged_json = None
    output = os.path.join(get_env('OUTPUT'))
    try:
        process_pool.shutdown(wait=True)
    except Exception as ex:
        process_pool.shutdown(wait=False, cancel_futures=True)
    if type(json_reports) is list:
        if scenario:
            json_reports = join_scenario_reports(json_reports)
        merged_json = join_feature_reports(json_reports)
    else:
        merged_json = json_reports
    if global_vars.progress_bar_instance:
        global_vars.progress_bar_instance.finish()
    status_info = os.path.join(output, global_vars.report_filenames['report_overall'])
    with open(status_info, 'w') as file_info:
        over_status = {'status': get_overall_status(merged_json)}
        file_info.write(json.dumps(over_status))
    path_info = os.path.join(output, global_vars.report_filenames['report_json'])
    with open(path_info, 'w') as file_info:
        file_info.write(json.dumps(merged_json))
    if get_param('dry_run'):
        print('Generating outputs...')
    generate_reports(merged_json)


def remove_temporary_files(parallel_processes, json_reports):
    """
    Remove temporary files created during the test execution.

    Args:
        parallel_processes (int): Number of parallel processes.
        json_reports (list): List of JSON reports.
    """
    path_info = os.path.join(
        os.path.join(get_env('OUTPUT'), global_vars.report_filenames['report_json'])
    )
    if os.path.exists(path_info):
        with open(path_info, 'r') as json_file:
            results_json = json.load(json_file)
            if 'features' and results_json['features']:
                return

    for i in range(parallel_processes):
        result_temp = os.path.join(gettempdir(), 'result{}.tmp'.format(i + 1))
        if os.path.exists(result_temp):
            try:
                os.remove(result_temp)
            except Exception as remove_ex:
                print(remove_ex)
        # Appending the process ID to the stdout file as a prefix, to avoid conflicts with other BehaveX processes
        path_stdout = os.path.join(gettempdir(), '{}_stdout{}.txt'.format(os.getpid(), i + 1))
        if os.path.exists(path_stdout):
            try:
                os.chmod(path_stdout, 511)  # nosec
                os.remove(path_stdout)
            except Exception as remove_ex:
                print(remove_ex)

    name = multiprocessing.current_process().name.split('-')[-1]
    stdout_file = os.path.join(gettempdir(), 'std{}2.txt'.format(name))
    logger = logging.getLogger()
    logger.propagate = False
    for handler in logging.root.handlers:
        logger.removeHandler(handler)
    for handler in logger.handlers:
        logger.removeHandler(handler)
    logging._handlers = []
    console_log = logging.StreamHandler(sys.stdout)
    console_log.setLevel(get_logging_level())
    logger.addHandler(console_log)
    if os.path.exists(stdout_file):
        os.chmod(stdout_file, 511)  # nosec
        if not os.access(stdout_file, os.W_OK):
            os.remove(stdout_file)
    # removing any pending temporary files
    if type(json_reports) is not list:
        json_reports = [json_reports]
    for json_report in json_reports:
        if 'features' in json_report and json_report['features']:
            feature_name = os.path.join(get_env('OUTPUT'), u'{}.tmp'.format(json_report['features'][0]['name']))
            if os.path.exists(feature_name):
                try:
                    os.remove(feature_name)
                except Exception as remove_ex:
                    print(remove_ex)


def processing_xml_feature(json_output, scenario_line, feature_filename,
                           scenarios_to_run_in_feature=None, lock=None,
                           shared_removed_scenarios=None):
    """
    Process the XML feature and update the JSON output.

    Args:
        json_output (dict): JSON output of the test execution.
        scenario (Scenario): Scenario object.
        feature_filename (str): Name of the feature file.
        scenarios_to_run_in_feature (int): Number of scenarios to run in the feature.
        lock (Lock): Multiprocessing lock.
        shared_removed_scenarios (dict): Shared dictionary of removed scenarios.
    """
    if lock:
        lock.acquire()
    try:
        feature_contains_scenarios = True if (json_output['features'] and
                                              'scenarios' in json_output['features'][0] and
                                              json_output['features'][0]["scenarios"]) else False
        if type(shared_removed_scenarios) == DictProxy and not feature_contains_scenarios:
            # Assuming the scenario was removed from the execution (using scenarios.remove) as it is not in execution
            if feature_filename not in shared_removed_scenarios:
                shared_removed_scenarios[feature_filename] = 1
            else:
                shared_removed_scenarios[feature_filename] += 1
        if feature_contains_scenarios:
            reported_scenarios = json_output['features'][0]['scenarios']
            executed_scenarios = []
            for reported_scenario in reported_scenarios:
                if reported_scenario['line'] == scenario_line:
                    executed_scenarios.append(reported_scenario)
            json_output['features'][0]['scenarios'] = executed_scenarios
            feature_name = os.path.join(
                get_env('OUTPUT'), u'{}.tmp'.format(os.path.basename(feature_filename))
            )
            processed_feature_data = json_output['features'][0]
            processed_feature_data['scenarios'] = executed_scenarios
            if os.path.exists(feature_name):
                for _ in range(0, 10):
                    try:
                        processed_feature_data = json.load(open(feature_name, 'r'))
                        with open(feature_name, 'w') as feature_file:
                            for executed_scenario in executed_scenarios:
                                processed_feature_data['scenarios'].append(executed_scenario)
                            json.dump(processed_feature_data, feature_file)
                        break
                    except Exception as ex:
                        logging.debug(ex)
                        logging.debug('Retrying reading from {}'.format(feature_name))
                        time.sleep(1)
            else:
                with codecs.open(feature_name, 'w', 'utf8') as feature_file:
                    json.dump(processed_feature_data, feature_file)
            # calculate the total number of scenarios that should be executed
            removed_scenarios = 0
            if shared_removed_scenarios and feature_filename in shared_removed_scenarios:
                removed_scenarios = shared_removed_scenarios[feature_filename]
            if scenarios_to_run_in_feature is None:
                total_scenarios = len_scenarios(feature_filename) - removed_scenarios
            else:
                total_scenarios = scenarios_to_run_in_feature - removed_scenarios
            if len(processed_feature_data['scenarios']) == total_scenarios:
                try:
                    report_xml.export_feature_to_xml(processed_feature_data, False)
                except Exception as ex:
                    traceback.print_exc()
                    print(ex)
                finally:
                    path_tmp = u'{}.tmp'.format(feature_name[:-4])
                    os.remove(path_tmp)
    except Exception as ex:
        raise ex
    finally:
        if lock:
            lock.release()


def _set_env_variables(args):
    """
    Set environment variables based on the given arguments.

    Args:
        args (Namespace): Parsed command-line arguments.
    """
    output_folder = os.path.normpath(get_env('output'))
    if os.path.isabs(output_folder):
        set_env_variable('OUTPUT', output_folder)
    else:
        set_env_variable('OUTPUT', os.path.abspath(output_folder))
    _store_tags_to_env_variable(args.tags)
    if get_param('include_paths'):
        set_env_variable('INCLUDE_PATHS', get_param('include_paths'))
    if get_param('include'):
        set_env_variable('INCLUDE', get_param('include'))
    if get_param('name'):
        set_env_variable('NAME', args.name)
    for arg in BEHAVEX_ARGS[4:]:
        set_env_variable(arg.upper(), get_param(arg))

    set_env_variable('TEMP', os.path.join(get_env('output'), 'temp'))
    set_env_variable('LOGS', os.path.join(get_env('output'), 'outputs', 'logs'))
    set_env_variable('LOGGING_LEVEL', get_logging_level())
    if platform.system() == 'Windows':
        set_env_variable('HOME', os.path.abspath('.\\'))
    set_env_variable('DRY_RUN', get_param('dry_run'))
    print_env_variables(
        [
            'HOME',
            'CONFIG',
            'OUTPUT',
            'TAGS',
            'PARALLEL_SCHEME',
            'PARALLEL_PROCESSES',
            'FEATURES_PATH',
            'TEMP',
            'LOGS',
            'LOGGING_LEVEL'
        ]
    )


def _store_tags_to_env_variable(tags):
    """
    Store tags to the environment variable.

    Args:
        tags (list): List of tags.
    """
    config = conf_mgr.get_config()
    tags_skip = config['test_run']['tags_to_skip']
    if isinstance(tags_skip, str) and tags_skip:
        tags_skip = [tags_skip]
    else:
        tags_skip = tags_skip
    tags = tags if tags is not None else []
    tags_skip = [tag for tag in tags_skip if tag not in tags]
    tags = tags + ['~@{0}'.format(tag) for tag in tags_skip] if tags else []
    if tags:
        for tag in tags:
            if get_env('TAGS'):
                set_env_variable('TAGS', get_env('tags') + ';' + tag)
            else:
                set_env_variable('TAGS', tag)
    else:
        set_env_variable('TAGS', '')


def _set_behave_arguments(features_path, multiprocess, execution_id=None, feature=None, scenario_line=None, config=None):
    """
    Set the arguments for Behave framework based on the given parameters.

    Args:
        features_path (str): Path to the features.
        multiprocess (bool): Whether to use multiprocessing.
        execution_id (str): Execution ID.
        feature (Feature): Feature object.
        scenario_line (int): Scenario line.
        config (ConfigRun): Configuration object.

    Returns:
        list: List of arguments to be used when executing Behave.
    """
    arguments = []
    output_folder = config.get_env('OUTPUT')
    if multiprocess:
        updated_features_path = features_path if not feature else feature
        updated_features_path = updated_features_path if not scenario_line else "{}:{}".format(updated_features_path, scenario_line)
        arguments.append(updated_features_path)
        arguments.append('--no-summary')
        worker_id = multiprocessing.current_process().name.split('-')[-1]

        arguments.append('--outfile')
        # Appending the process ID to the stdout file as a prefix, to avoid conflicts with other BehaveX processes
        arguments.append(os.path.join(gettempdir(), '{}_stdout{}.txt'.format(os.getpid(), worker_id)))

        arguments.append('-D')
        arguments.append(f'worker_id={worker_id}')
    else:
        if type(features_path) is list:
            for feature_path in features_path:
                arguments.append(feature_path)
        else:
            arguments.append(features_path)
        if get_param('dry_run'):
            arguments.append('--no-summary')
        else:
            arguments.append('--summary')
        arguments.append('--junit-directory')
        arguments.append(output_folder)
        arguments.append('--outfile')
        arguments.append(os.path.join(output_folder, 'behave', 'behave.log'))
        arguments.append('-D')
        arguments.append(f'worker_id=0')
    arguments.append('--no-skipped')
    arguments.append('--no-junit')
    run_wip_tests = False
    if get_env('tags'):
        tags = get_env('tags').split(';')
        for tag in tags:
            arguments.append('--tags')
            arguments.append(tag)
            if tag.upper() in ['WIP', '@WIP']:
                run_wip_tests = True
    if not run_wip_tests:
        arguments.append('--tags')
        arguments.append('~@WIP')
    arguments.append('--tags')
    arguments.append('~@MANUAL')
    args_sys = config.args
    set_args_captures(arguments, args_sys)
    if args_sys.no_snippets:
        arguments.append('--no-snippets')
    for arg in BEHAVE_ARGS:
        value_arg = getattr(args_sys, arg) if hasattr(args_sys, arg) else False
        if arg == 'include':
            if multiprocess or not value_arg:
                continue
            else:
                features_path = os.path.abspath(os.environ['FEATURES_PATH'])
                value_arg = value_arg.replace(features_path, 'features').replace(
                    '\\', '\\\\'
                )
        if arg == 'define' and value_arg:
            for key_value in value_arg:
                arguments.append('--define')
                arguments.append(key_value)
        if value_arg and arg not in BEHAVEX_ARGS and arg != 'define':
            arguments.append('--{}'.format(arg.replace('_', '-')))
            if value_arg and not isinstance(value_arg, bool):
                arguments.append(value_arg)
                if arguments == 'logging_level':
                    set_env_variable(arg, value_arg)
                else:
                    os.environ[arg] = str(value_arg)
    return arguments


def _get_feature_json_skeleton(behave_element):
    """
    Get the JSON skeleton for the given feature or scenario.

    Args:
        behave_element (Feature or Scenario): Behave element.

    Returns:
        str: JSON skeleton of the feature.
    """
    if type(behave_element) is Feature:
        feature = behave_element
    elif type(behave_element) is Scenario:
        feature = copy.copy(behave_element.feature)
        feature.scenarios = [behave_element]
    else:
        raise Exception("No feature or scenario to process...")
    execution_info = generate_execution_info([feature])
    return json.dumps(execution_info[0]) if execution_info else {}


def _get_progress_bar_instance(parallel_scheme, total_elements):
    """
    Get the progress bar instance.

    Args:
        parallel_scheme (str): Parallel scheme (feature or scenario).
        total_elements (int): Total number of elements.

    Returns:
        ProgressBar: Progress bar instance.
    """
    try:
        config = conf_mgr.get_config()
        print_updates_in_new_lines = True if config['progress_bar']['print_updates_in_new_lines'] else False
        progress_bar_prefix = "Executed {}s".format(parallel_scheme)
        progress_bar_instance = ProgressBar(progress_bar_prefix,
                                            total_elements,
                                            print_updates_in_new_lines=print_updates_in_new_lines)
    except Exception as ex:
        print("There was an error creating the progress bar: {}".format(ex))
        return None
    return progress_bar_instance


def set_args_captures(args, args_sys):
    """
    Set the capture arguments for Behave.

    Args:
        args (list): List of arguments.
        args_sys (Namespace): Parsed command-line arguments.
    """
    for default_arg in ['capture', 'capture_stderr', 'logcapture']:
        if not getattr(args_sys, 'no_{}'.format(default_arg)):
            args.append('--no-{}'.format(default_arg.replace('_', '-')))


def dump_json_results():
    """
    Dump the JSON results of the test execution.

    Returns:
        dict: JSON output of the test execution.
    """
    if multiprocessing.current_process().name == 'MainProcess':
        path_info = os.path.join(
            os.path.abspath(get_env('OUTPUT')),
            global_vars.report_filenames['report_json'],
        )
    else:
        process_name = multiprocessing.current_process().name.split('-')[-1]
        path_info = os.path.join(gettempdir(), 'result{}.tmp'.format(process_name))

    def _load_json():
        """this function load from file"""
        with open(path_info, 'r') as info_file:
            json_output_file = info_file.read()
            json_output_converted = json.loads(json_output_file)
        return json_output_converted

    json_output = {'environment': [], 'features': [], 'steps_definition': []}
    if os.path.exists(path_info):
        json_output = retry_file_operation(path_info, _load_json, return_value=True)
    return json_output


if __name__ == '__main__':
    main()




def after_scenario(context, scenario):
    if hasattr(context, 'new_scenario_name'):
        scenario.name = context.new_scenario_name
    if hasattr(context, 'new_feature_name'):
        scenario.feature.name = context.new_feature_name



# -*- coding: utf-8 -*-
"""
/*
* BehaveX - Agile test wrapper on top of Behave (BDD)
*/

Test report utility methods for retrieving summarized test execution information
 from features and scenarios
"""
# pylint: disable=W0703
# pylint: disable=W0703 , E1101

# __future__ is added in order to maintain compatibility
from __future__ import absolute_import, print_function

import codecs
import hashlib
import os
import re
import shutil
import string
import sys
import traceback
import unicodedata
from operator import getitem

from behave.model_core import Status

from behavex.conf_mgr import get_env, get_param, set_env
from behavex.global_vars import global_vars


def gather_steps_with_definition(features, steps_definition):
    all_steps = []
    for feature in features:
        for scenario in feature['scenarios']:
            all_steps += scenario['steps']
            if scenario['background']:
                all_steps += scenario['background']['steps']
    result = {}
    if steps_definition:
        for id_hash, definition in steps_definition.items():
            result[definition] = {'steps': []}
            for step in all_steps:
                if step['hash'] == int(id_hash):
                    result[definition]['steps'].append(step)
        for definition, value in result.items():
            result[definition] = get_summary_definition(value['steps'])

    return result


def get_summary_definition(steps):
    result = {'steps': {}, 'status': []}
    for step in steps:
        result['status'].append(step['status'])
        execution = 1 if step['status'] in ('failed', 'passed') else 0
        if step['name'] in result['steps']:
            result['steps'][step['name']]['executions'] += execution
            result['steps'][step['name']]['time'] += step['duration']
            result['steps'][step['name']]['status'].append(step['status'])
            result['steps'][step['name']]['appearances'] += 1
        else:
            result['steps'][step['name']] = {
                'executions': execution,
                'time': step['duration'],
                'status': [step['status']],
                'appearances': 1,
            }
    total_time = 0.0
    executions = 0
    appearances = 0
    any_status_failed = False
    any_status_passed = False
    for step_instanced in result['steps'].values():
        total_time += step_instanced['time']
        executions += step_instanced['executions']
        appearances += step_instanced['appearances']
        if 'passed' in step_instanced['status']:
            any_status_passed = True
        elif 'failed' in step_instanced['status']:
            any_status_failed = True
    if any_status_failed:
        result['overall_status'] = 'failed'
    elif any_status_passed:
        result['overall_status'] = 'passed'
    else:
        result['overall_status'] = 'skipped'
    if executions > 0:
        avg_time = total_time / executions
    else:
        avg_time = 0
    result['avg'] = avg_time
    result['time'] = total_time
    result['executions'] = executions
    result['appearances'] = appearances
    return result


def calculate_status(list_status):
    set_status = set(list_status)
    if 'untested' in set_status:
        set_status.remove('untested')
        set_status.add('skipped')
    if 'undefined' in set_status:
        set_status.remove('undefined')
        set_status.add('skipped')
    if 'failed' in set_status:
        return 'failed'
    elif {'skipped'} == set_status:
        return 'skipped'
    else:
        return 'passed'


def gather_steps(features):
    steps = {}
    for feature in features:
        for scenario in feature['scenarios']:
            steps_back = [step for step in scenario['background']['steps']]
            for step in scenario['steps'] + steps_back:
                if not step['name'] in steps:
                    steps[step['name']] = {
                        'quantity': 0,
                        'total_duration': 0,
                        'appearances': 0,
                        'passed': 0,
                        'failed': 0,
                        'skipped': 0,
                    }
                steps[step['name']]['appearances'] += 1

                add = (
                    0
                    if step['status'].lower() in ['skipped', 'untested', 'undefined']
                    else 1
                )
                passed = 1 if step['status'] == 'passed' else 0
                failed = 1 if step['status'] == 'failed' else 0
                steps[step['name']]['passed'] += passed
                steps[step['name']]['failed'] += failed
                steps[step['name']]['quantity'] += add
                steps[step['name']]['skipped'] += not add
                steps[step['name']]['total_duration'] += step['duration']

    return steps


def gather_errors(scenario, retrieve_step_name=False):
    error_msg = list(map(lambda line: strip_ansi_codes(line), scenario['error_msg']))
    if retrieve_step_name:
        return error_msg, scenario['error_lines'], scenario['error_step']
    else:
        return error_msg, scenario['error_lines']


def pretty_print_time(seconds_float, sec_decimals=1):
    seconds_float = round(seconds_float, sec_decimals)
    hours, remainder = divmod(seconds_float, 3600)
    minutes, seconds = divmod(remainder, 60)
    seconds = (
        int(seconds) if float(seconds).is_integer() else round(seconds, sec_decimals)
    )
    def _pretty_format(cant, unit):
        return f'{cant}{unit}' if cant > 0 else ''
    time_string = f'{_pretty_format(int(hours), "h")} {_pretty_format(int(minutes), "m")} {_pretty_format(seconds, "s")}'
    return "0s" if time_string.strip() == '' else time_string


def normalize_path(path):
    exp = re.compile('\\\\|/')
    return os.path.join(*exp.split(path))


def resolving_type(step, scenario, background=True):
    if step['index'] == 0:
        return step['step_type'].title()
    else:
        steps = scenario['steps']
        if 'background' in list(step.keys()) and not background:
            steps = scenario['background']['steps']
        # noinspection PyBroadException
        try:
            previous_type = steps[step['index'] - 1]['step_type']
        except BaseException:
            previous_type = step['step_type']
        if previous_type == step['step_type']:
            return 'And'
        else:
            return step['step_type'].title()


def retry_file_operation(dest_path, execution, return_value=False):
    """Executes a file operation with retry logic for handling file access conflicts.

    This function attempts to execute a file operation (like copy, delete, etc.) and
    provides an interactive retry mechanism if the files are locked by another process.

    Args:
        dest_path (str): The file system path where the operation will be performed
        operation (callable): A function containing the file operation to execute
        return_value (bool, optional): Whether to return the result of the operation. Defaults to False.

    Returns:
        Any: The result of the operation if return_value is True, otherwise None

    Example:
        >>> retry_file_operation('/path/to/dir', lambda: shutil.rmtree('/path/to/dir'))
    """
    passed = False
    following = True
    retry_number = 1
    while not passed and following:
        try:
            result = execution()
            passed = True
            if return_value:
                return result
        except Exception as exception:
            menu = (
                '\nAnother program is using the files in path {}. '
                "Please, close the program and press 'c' to continue "
                " or 'a' for abort. Retry number {}\n\n.{}"
            )
            option_correct = False
            retry_number += 1
            option = (
                str(
                    input(
                        menu.format(os.path.abspath(dest_path), retry_number, exception)
                    )
                )
                .upper()
                .strip()
            )
            while not option_correct and not passed:
                if option == 'C':
                    following = True
                    option_correct = True
                elif option == 'A':
                    exit()
                else:
                    following = True
                    option_correct = False
                    msg = (
                        "\nInvalid option. Please, press 'c' to continue "
                        "or 'a' to abort\n"
                    )
                    option = str(input(msg)).upper().strip()


def get_status(dictionary):
    return (
        dictionary.get('failed', False)
        or dictionary.get('passed', False)
        or dictionary.get('skipped', 'skipped')
    )


def get_test_execution_tags():
    if not get_env('behave_tags'):
        behave_tags_path = os.path.join(
            os.path.abspath(get_env('OUTPUT')), global_vars.behave_tags_file
        )
        behave_tags_file = open(behave_tags_path, 'r')
        behave_tags = behave_tags_file.readline().strip()
        behave_tags_file.close()
        os.chmod(behave_tags_path, 511)  # nosec
        set_env('behave_tags', behave_tags)
        return get_env('behave_tags')
    else:
        return get_env('behave_tags')


def match_for_execution(tags):
    # Check filter
    tag_re = re.compile(r'@?[\w\d\-_.]+')
    tags_filter = get_test_execution_tags()
    # Eliminate tags put for param dry-rFun
    if get_param('dry_run'):
        if 'BHX_MANUAL_DRY_RUN' in tags:
            tags.remove('BHX_MANUAL_DRY_RUN')
        if 'MANUAL' in tags:
            tags.remove('MANUAL')
    # Set scenario tags in filter
    for tag in tags:
        # Try with and without @
        def replace(arroba, x, y):
            return re.sub(r'^{}{}$'.format(arroba, x), 'True', y)

        tags_filter = ' '.join(
            [replace('@', tag, tag_) for tag_ in tags_filter.split()]
        )

        tags_filter = ' '.join([replace('', tag, tag_) for tag_ in tags_filter.split()])

    # Set all other tags to False
    for tag in tag_re.findall(tags_filter):
        if tag not in ('not', 'and', 'or', 'True', 'False'):
            tags_filter = tags_filter.replace(tag + ' ', 'False ')
    return tags_filter == '' or eval(tags_filter)  # nosec


def copy_bootstrap_html_generator(output):
    dest_path = os.path.join(output, 'outputs', 'bootstrap')
    bootstrap_path = ['outputs', 'bootstrap']
    bootstrap_path = os.path.join(global_vars.execution_path, *bootstrap_path)
    if os.path.exists(dest_path):
        retry_file_operation(dest_path, lambda: shutil.rmtree(dest_path))
    retry_file_operation(
        dest_path, lambda: shutil.copytree(bootstrap_path, dest_path)
    )


def get_overall_status(output):
    if not output:
        return {'status': 'skipped'}
    overall_status = {
        feature['status']: feature['status'] for feature in output['features']
    }
    overall_status_end = get_status(overall_status)
    return overall_status_end


def get_save_function(path, content):
    def fun_save():
        with codecs.open(path, 'w', 'utf8') as file_:
            file_.write(content)
        os.chmod(path, 511)  # nosec

    return fun_save


def create_log_path(name, execution_retry=False):
    name = get_string_hash(name)
    if sys.version_info < (3, 0):
        path = os.path.join(get_env('logs').decode('utf8'), str(name))
    else:
        path = os.path.join(get_env('logs'), str(name))
    initial_path = path
    scenario_outline_index = 1
    while os.path.exists(path):
        if execution_retry:
            return path
        path = initial_path + u'_' + str(scenario_outline_index)
        scenario_outline_index += 1
    try:
        os.makedirs(path)
    except FileExistsError:
        # path already exists
        pass
    return path


def get_error_message(error_message):
    if not error_message:
        return u''
    if isinstance(error_message, Exception):
        if hasattr(error_message, 'message'):
            # noinspection PyBroadException
            try:
                error_message = traceback.format_tb(error_message.message)
            except BaseException:
                error_message = repr(error_message.message)
        if hasattr(error_message, 'exc_traceback'):
            error_message = traceback.format_tb(error_message.exc_traceback)
    elif not isinstance(error_message, str):
        error_message = repr(error_message)
    formatted_error = u'\n'.join([16 * u' ' + line for line in text(error_message).split('\n')]).strip()
    return formatted_error


def text(value, encoding=None, errors=None):
    if value:
        value = value.name if isinstance(value, Status) else value.replace('\\', '/')
    if encoding is None:
        encoding = 'unicode-escape'
    if errors is None:
        errors = 'replace'

    if isinstance(value, str):
        # -- PASS-THROUGH UNICODE (efficiency):
        return value
    elif isinstance(value, bytes):
        return str(value, encoding, errors)
    elif isinstance(value, bytes):
        # -- MAYBE: filename, path, etc.
        try:
            return value.decode(sys.getfilesystemencoding())
        except UnicodeError:
            return value.decode('utf-8', 'replace')
    else:
        # -- CONVERT OBJECT TO TEXT:
        try:
            output_text = str(value)
        except UnicodeError as e:
            output_text = str(e)
        return output_text


def get_string_hash(string_to_hash):
    return str(hashlib.sha256(string_to_hash.encode('UTF-8')).hexdigest())


def normalize_filename(input_name):
    valid_name = None
    try:
        if sys.version_info[0] == 2:
            if isinstance(input_name, str):
                if isinstance(input_name, str):
                    input_name = text(input_name)
            else:
                raise TypeError
            if not input_name:
                return u''
            # xrange has been replaced for range
            characters = ''.join(map(chr, range(255)))
            reserved_characters = '<>:\'"/\\|?*'
            for character in reserved_characters:
                characters = characters.replace(character, '')
            valid_name = ''.join(char for char in input_name if char in characters)
            valid_name = unicodedata.normalize('NFKD', valid_name).encode(
                'ascii', 'ignore'
            )
            valid_name = string.capwords(valid_name)
            valid_name = valid_name.replace(' ', '_')
            if isinstance(valid_name, str):
                valid_name = text(valid_name)
        if sys.version_info[0] == 3:
            if isinstance(input_name, str):
                pass
            else:
                raise TypeError
            characters = ''.join(chr(i) for i in range(255))
            reserved_characters = '<>:\'"/\\|?*'
            for character in reserved_characters:
                characters = characters.replace(character, '')
            valid_name = ''.join(char for char in input_name if char in characters)
            valid_name = unicodedata.normalize('NFKD', valid_name).encode(
                'ascii', 'ignore'
            )
            if isinstance(valid_name, str):
                valid_name = string.capwords(valid_name)
            else:
                valid_name = bytes.title(valid_name)
            if isinstance(valid_name, str):
                valid_name = valid_name.replace(' ', '_')
            else:
                valid_name = valid_name.replace(b' ', b'_')
            if isinstance(valid_name, str):
                valid_name = bytes(valid_name, encoding='UTF-8')
        if isinstance(valid_name, bytes):
            valid_name = valid_name.decode('utf8')
        return valid_name
    except TypeError:
        err_msg = 'Input should be str or unicode, but received a: {}'.format(
            type(input_name)
        )
        print(err_msg)
        return


def get_environment_details():
    """ Get environment information from the ENVIRONMENT_DETAILS environment variable.
    The environment information is a comma-separated string that is split into a list."""
    environment_details_raw_data = os.getenv('ENVIRONMENT_DETAILS', None)
    environment_details = environment_details_raw_data.split(',') if environment_details_raw_data else []
    return environment_details

def strip_ansi_codes(from_str: str):
    return re.sub(r'\x1B\[[0-?9;]*[mGJK]', '', from_str)



import logging

from behave import given, then, when


@given('a failing condition')
def step_impl(context):
    context.condition = 'fail'
    logging.info('a failing condition')

@given('a passing condition')
def step_impl(context):
    context.condition = 'pass'
    logging.info('a passing condition')

@given('a condition to skip the scenario')
def step_impl(context):
    context.condition = 'skip'
    logging.info('a condition to skip the scenario')

@given('a condition to exit the scenario')
def step_impl(context):
    context.condition = 'exit'
    logging.info('a condition to exit the scenario')

@given('a condition to leave the scenario untested')
def step_impl(context):
    context.condition = 'untested'
    logging.info('a condition to leave the scenario untested')

@then('I perform the condition')
def step_impl(context):
    if context.condition == 'fail':
        # This step will cause the test to fail
        assert False, "This step is designed to fail"
    elif context.condition == 'pass':
        # This step will pass
        assert True
    elif context.condition == 'skip':
        # This step will be skipped
        context.scenario.skip("This scenario is skipped")
    elif context.condition == 'exit':
        # This step will be skipped
        exit(1)
    elif context.condition == 'untested':
        # This step will be skipped
        pass

@given('I rename the {feature_or_scenario} from context to have the suffix "{suffix}"')
def step_impl(context, feature_or_scenario, suffix):
    if feature_or_scenario == 'feature':
        context.new_feature_name = context.feature.name + suffix
        logging.info('I rename the feature from \n"{}" \nto \n"{}"'.format(context.feature.name, context.new_feature_name))
    elif feature_or_scenario == 'scenario':
        context.new_scenario_name = context.scenario.name + suffix
        logging.info('I rename the scenario from \n"{}" \nto \n"{}"'.format(context.scenario.name, context.new_scenario_name))
    else:
        raise ValueError('Invalid element, it should be "feature" or "scenario"')



# -*- coding: utf-8 -*-
"""

/*
* BehaveX - Agile test wrapper on top of Behave (BDD)
*/

Parse all framework arguments, the ones from BehaveX and
the ones from Behave supported by the wrapper
"""
# __future__ has been added to maintain compatibility
from __future__ import absolute_import

import argparse

BEHAVE_ARGS = [
    'no_color',
    'color',
    'define',
    'exclude',
    'include',
    'no_junit',
    'junit',
    'junit_directory',
    'steps_catalog',
    'no_skipped',
    'show_skipped',
    'lang',
    'no_snippets',
    'snippets',
    'no_multiline',
    'multiline',
    'no_capture',
    'name',
    'capture',
    'no_capture_stderr',
    'capture_stderr',
    'no_logcapture',
    'logcapture',
    'logging_level',
    'logging_format',
    'logging_datefmt',
    'logging_filter',
    'logging_clear_handlers',
    'no_summary',
    'summary',
    'outfile',
    'quiet',
    'no_source',
    'show_source',
    'stage',
    'stop',
    'tags',
    'no_timings',
    'show_timings',
    'verbose',
    'wip',
    'expand',
    'lang_list',
    'lang_help',
    'tags_help',
]

BEHAVEX_ARGS = [
    'output_folder',
    'config',
    'dry_run',
    'tags',
    'parallel_scheme',
    'parallel_processes',
    'show_progress_bar',
]


def parse_arguments(args):
    """Process all command line arguments"""
    parser = argparse.ArgumentParser(
        description='BehaveX - test automation wrapper on top of Behave'
    )
    parser.add_argument("paths", nargs="*",
                        help="Features path")
    parser.add_argument(
        '-c',
        '--config',
        help='BehaveX configuration file to use by default.',
        required=False,
    )
    parser.add_argument(
        '-t',
        '--tags',
        action='append',
        help='Tags used to properly filter the tests to run. \
                                When multiple --tags (-t) arguments are \
                                provided it means a logical AND \
                                (e.g. -t @TAG_1 -t @TAG_2 means \
                                @TAG_1 AND @TAG_2). \
                                When multiple comma separated tags are \
                                provided as part of the same --tags (-t) \
                                argument it means a logical OR \
                                (e.g. -t @TAG_1,@TAG_2 means \
                                @TAG_1 OR @TAG_2)',
        required=False,
    )
    parser.add_argument(
        '-o',
        '--output-folder',
        default='',
        help='Specifies the output folder path where the test report will be stored. '
             'The path can be specified as a relative path. The value of this argument '
             'can be accessed in testing implementations using the "OUTPUT" environment variable',
        required=False,
    )
    parser.add_argument(
        '-d',
        '--dry-run',
        default='',
        action='store_true',
        help='Invokes formatters without executing the steps.',
        required=False,
    )

    # ------------------- Behave arguments -------------------#
    parser.add_argument(
        '--no-color',
        '--no_color',
        help='Disable the use of ANSI color escapes.',
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '--color',
        help='Use ANSI color escapes. This is the default '
        'behaviour. This switch is used to override a '
        'configuration file setting.',
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '-D',
        '--define',
        help='Define user-specific data in config.userdata '
        'dictionary. Example: -D foo=bar to store it in '
        "config.userdata['foo'].",
        action='append',
        required=False,
    )
    parser.add_argument(
        '--exclude',
        help="Don't run feature files matching regular expression PATTERN.",
        required=False,
    )
    parser.add_argument(
        '-i',
        '--include',
        help='Only run feature files matching regular expression PATTERN.',
        required=False,
    )
    parser.add_argument(
        '--name',
        help='Execute feature elements matching a part of the given name. '
             'If this option is specified more than once, '
             'it will match against all given names.',
        required=False,
    )
    parser.add_argument(
        '--no-capture',
        '--no_capture',
        help="Don't capture stdout (any stdout output will be printed immediately.)",
        default=False,
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '--capture',
        help='Capture stdout (any stdout output will be '
        'printed if there is a failure.) This is the '
        'default behaviour. This switch is used to '
        'override a configuration file setting.',
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '--no-capture-stderr',
        '--no_capture_stderr',
        help="Don't capture stderr (any stderr output will be printed immediately.)",
        default=False,
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '--capture-stderr',
        '--capture_stderr',
        help='Capture stderr (any stderr output will be pri'
        'nted if there is a failure) This is the default'
        ' behaviour. This switch is used to override a '
        'configuration file setting.',
        default=False,
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '--no-logcapture',
        '--no_logcapture',
        help="Don't capture logging. Logging configuration will be left intact.",
        action='store_true',
        default=False,
        required=False,
    )
    parser.add_argument(
        '--log-capture',
        '--log_capture',
        help='Capture logging. All logging during a step will'
        ' be captured and displayed in the event of a '
        'failure. This is the default behaviour. This '
        'switch is used to override a configuration file'
        ' setting.',
        action='store_true',
        required=False,
    )
    # parser.add_argument('--summary',
    #                     help="Display the summary at the end of the run.",
    #                     action="store_true",
    #                     required=False)
    # parser.add_argument('-q',
    #                     '--quiet',
    #                     help="Alias for --no-snippets --no-source.",
    #                     action="store_true",
    #                     required=False)
    # parser.add_argument('-s',
    #                     '--no-source',
    #                     help="Don't print the file and line of the step d"
    #                          "efinition with the steps.",
    #                     action="store_true",
    #                     required=False)
    # parser.add_argument('--show-source',
    #                     help="Print the file and line of the step definition "
    #                          "with the steps. This is the default behaviour."
    #                          "This switch is used to override a configuration"
    #                          " file setting.",
    #                     action='store_true',
    #                     required=False)
    parser.add_argument(
        '-ns',
        '--no-snippets',
        '--no_snippets',
        help="Don't print snippets for unimplemented steps.",
        default=False,
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '--stop',
        help='Specifies that the test execution should stop at the first failure. '
             'Note that this argument is not supported for parallel test executions',
        action='store_true',
        required=False,
    )
    parser.add_argument(
        '--tags-help',
        '--tags_help',
        help='Show help for tag expressions.',
        action='store_true',
        required=False,
    )
    # -------------------- adding lastly ---------------------------------------
    parser.add_argument(
        '--logging-level',
        '--logging_level',
        default='INFO',
        choices=['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'],
        help='Specifies the logging level to capture.',
        required=False,
    )
    parser.add_argument(
        '--parallel-processes',
        default=1,
        type=int,
        help='Specifies the number of parallel processes that can be executed simultaneously.',
        required=False,
    )
    parser.add_argument(
        '--parallel-scheme',
        choices=['feature', 'scenario'],
        default='scenario',
        help="Specifies whether parallel execution should be performed at the scenario or feature level.",
        required=False,
    )
    parser.add_argument(
        '--parallel-delay',
        type=int,
        default=0,
        help='Delay in milliseconds before starting each parallel process')
    parser.add_argument(
        '-ip',
        '--include-paths',
        default=[],
        nargs='*',
        help='Filters the test set to the specified list of features.'
        'or feature file locations (FEATURE_FILE:LINE).',
    )
    parser.add_argument(
        '-rf',
        '--rerun-failures',
        help='Enables re-execution of the failing scenarios. '
             'After each execution, the failing scenarios are saved '
             'in a "failing_scenarios.txt" file in the output folder. '
             'This argument specifies the location of the "failing_scenarios.txt" file. '
             'For example: --rf ./output/failing_scenarios.txt',
        required=False
    )
    parser.add_argument(
        '-spb',
        '--show-progress-bar',
        help="Shows the execution progress bar in console.",
        default=False,
        action='store_true',
        required=False,
    )

    # parser.add_argument('--logging-format',
    #                    help="Specify custom format to print statements. Uses "
    #                          " the same format as used by standard logging "
    #                        "handlers. The default is (%levelname)s:w%(name)s:"
    #                          "%(message)s%.",
    #                     required=False)
    # parser.add_argument('--logging-datefmt',
    #                     help="Specify custom date/time format to print "
    #                          "statements. Uses the same format as used by "
    #                          "standard logging handlers.",
    #                     required=False)
    #
    return parser.parse_args(args)



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='behavex',
    version='4.1.0',
    license="MIT",
    platforms=['any'],
    python_requires='>=3.5',
    author='Hernan Rey',
    author_email='behavex_users@googlegroups.com',
    url='https://github.com/hrcorval/behavex',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    description='Agile testing framework on top of Behave (BDD).',
    long_description_content_type='text/markdown',
    long_description=long_description,
    entry_points={
        'console_scripts': [
            'behavex = behavex.runner:main',
        ],
    },
    install_requires=[
        'behave==1.2.6',
        'behavex-images',
        'jinja2',
        'configobj',
        'htmlmin',
        'csscompressor'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Testing',
    ],
)
